package Sophos::Reports::Report;
use strict;
use warnings;

use Carp;
use Text::CSV_XS;
use CGI qw(Tr td start_table end_table div img span);
use Date::Parse qw(str2time);
use GD::Graph::colour qw(_rgb);

use PureMessage::Version qw($PREFIX);

use Sophos::Reports::TimePeriods qw(get_day get_week get_adjustment);
use Sophos::Reports::Parameters qw(parameters_to_hash);

use Sophos::Reports::Graphic::Lines;
use Sophos::Reports::Graphic::GradientBars;

my $DAY_IN_SECONDS = 86400;
my $MONTH_IN_SECONDS= 86400 * 31;
my $YEAR_IN_SECONDS = 86400 * 365;

return 1;

sub new {
    my ($proto, %opts) = @_;
    my $class = ref($proto) || $proto;

    # Silly perl thinks we're commenting when we specify our colours
    no warnings 'qw';

    my $self = {
	graph => {
	    labelclr      => '#003151',
	    axislabelclr  => '#003151',
	    valuesclr     => '#003151',
	    textclr       => '#003151',
	    bgclr         => '#e3e3e3',
	    fgclr         => '#8ea2ae',

	    transparent    => 1,

	    t_margin      => 10,
	    b_margin      => 8,
	    l_margin      => 2,
	    r_margin      => 15,

        # bar graph color palettes from
        # http://ca-uiguide.eng.sophos/web/color_ui.html
        dclrs => [qw(
            #3A5081 #F7A724 #C9B8D3 #DD442D #FFFF8F #A3C978 #407F76 #313131
            #87CEEB #FFA07A #EE82EE #CD5C5C #F0E68C #98FB98 #8FBC8F #DCDCDC
            #000080 #FF4500 #800080 #8b0000 #FFFF00 #808000 #005252 #949494
        )],
	},
	lines => {
	    y_min_value   => 0,
	    line_width    => 3,
	    y_long_ticks  =>  1,
	    x_label_skip  => 1,
	    x_tick_number => undef,
	},
	bars => {
	    y_min_value   => 0,
	    borderclrs    => [undef],
	    cumulate      => 1,
	    bar_spacing   => 25,
	    x_tick_number => undef,
	    x_label_skip  => 1,
	},
	pie => {
	    transparent   => 0,
	    pie_placement => 'w',
            height => 260,
            width => 575,
	},
	nodata => 'report_nodata',

	font_size => 7,
	x_axis_font_size => 7,
        height => 260,
        width => 575,

	title => "/img-lib/en/Template/ttl_blank_lrg.jpg",
        nodata_dir => 'SUBCLASS SHOULD DEFINE',
        nodata_path => 'SUBCLASS SHOULD DEFINE',

	column_title => '',
        formatted_dates => 0,
    };

    bless $self, $class;

    if (exists($opts{report_uri})) {
        # When instantiated from ReportFactory
        $self->_init_from_uri($opts{report_uri});
    }

    # Add options to this object's attributes
    $self->{$_} = $opts{$_} for keys %opts;

    $self->set_small if $self->{small};

    return $self;
}

sub _init_from_uri {
    my ($self, $uri) = @_;

    croak "uri is required" unless defined $uri;

    my ($parameters, $output) = $uri =~ m#^([^.?]+)(?:\.([^/]*))?#
	or die "Invalid parameters";

    die "No parameters" unless $parameters;

    $self->{output} = $output || 'html';

    my @split = split(m{/}, $parameters);

    $self->{type} = shift @split || die "No report type specified";
    $self->{date} = shift @split || die "No report date specified";
    $self->{span} = shift @split || die "No report span specified";

     if ($self->{span} =~ /sfn=(\d)/) {
         $self->{sfn} = $1;
         $self->{span} =~ s/\s*sfn=\d//;
     }

    die "Too many parameters" if @split;

    $self->{small} = $self->{type} =~ s/^small//i;
}

sub report_id {
    my ($self) = @_;
    my ($report_id) = ref($self) =~ /^Sophos::Reports::Report::(.*)$/;
    return $report_id;
}

sub set_small {
    my ($self) = @_;
    $self->{pie}{height} = 125;
    $self->{pie}{width} = 130;
    $self->{pie}{pie_width} = 110;
    $self->{pie}{pie_height} = 110;
    $self->{x_axis_font_size} = 5;
    $self->{height} = 125;
    $self->{width} = 275;
}

sub dbh {
    die "Must be implemented by subclass";
}

sub clean_data {
    my ($self, $data, %args) = @_;
    
    my $type = $args{type} || $self->{type};
    my $rows = $args{rows} || 0;
    $args{cols} ||= 0;

    my $has_data = 0;

    for my $i (0 .. $rows-1) {
	$data->[$i] ||= [];
	my $cols = $type =~ /^(?:table|pie)$/ ? $args{cols} : @{$data->[$i]};

	for my $j (0 .. $cols-1) {
	    $data->[$i][$j] ||= 0;
	    $has_data ||= $data->[$i][$j];
	}
    }

    return $has_data;
}

sub make_uri {
    my ($self,%opts) = @_;

    my $old_args = $self->{args} || {};
    my $new_args = $opts{args} || {};
    my %args = ( %$old_args, %$new_args );

    my $uri;

    my %prms = (parameters_to_hash($self->{uri}), %opts);

    $uri = "/Report/$prms{report}/$prms{type}/$prms{date}/$prms{span}";
    $uri .= ".$prms{output}" if $prms{output};
    if (%args) {
	$uri .= '?' . join('&amp;',map {"$_=$args{$_}"} keys %args);
    }

    return $uri;
}

sub pretty_filename {
    my ($self,$type) = @_;
    die "type required" unless $type;
    return "$self->{report}-$self->{date}$self->{span}.$type";
}

sub set_y_max {
    my ($self, $data) = @_;
    return unless ref $data eq 'ARRAY';
    return unless $self->{type};
    
    my $lines = $self->{type} eq 'lines' ? 1 : $self->{type} eq 'bars' ? 0 : return;

    my $max = 0;
    for my $row (@$data) {
	my $row_max = 0;
	for my $col (@$row) {
	    next unless defined $col;
	    if ($lines) {
		$row_max = $col > $row_max ? $col : $row_max;
	    }
	    else {
		$row_max += $col;
	    }
	}
	$max = $row_max > $max ? $row_max : $max;
    }

    # Set y_max to 1.25 times the max value
    my $ymax = $max;
    $ymax = int(2*$ymax);
    substr($ymax,$_,1) = 0 for int(length($ymax)/2) .. length($ymax)-1;
    $self->{graph}{y_max_value} = $ymax;
}

sub csv {
    my $self = shift;
    my $csv;
    
    my $proc = Text::CSV_XS->new;

    my $output = $self->{output} || '';
    my $data = $output eq 'dat' ? $self->data : $self->csv_data;

    my @rv;
    foreach my $i (@$data) {
        $proc->combine(@$i);
        push(@rv, $proc->string);
    }
    $csv = join("\n", @rv, '');
    
    return $csv;
}

sub sql {
    my $self = shift;
    return $self->data(get_sql => 1);
}

sub time_period {
    my $self = shift;
    my $date = $self->{date} || die "Must set date!";
    my $span = $self->{span} || die "Must set span!";
    tr/'\\//d for ($date, $span);

    if ($span =~ /^-/ && $self->{sfn}) {
	return ("'$date'::timestamptz + '$span'::interval", "'$date'::timestamptz")
    }
    elsif ($span =~ /^\d+\s*minutes$/) {
	return ("'$date'::timestamptz", "'$date'::timestamptz + '$span'::interval")
    }
    elsif ($span =~ /^today/) {
        return ("date_trunc('day',now())::timestamptz" , "'tomorrow'::timestamptz")
    }
    else {
        my $field = ($self->{span} =~ /month/) ? 'month' :
                    ($self->{span} =~ /day/) ? 'day' : 'hour';
        return ("date_trunc('$field',now())::timestamptz + '" . $span . "'::interval",
                    "date_trunc('$field', now())::timestamptz")
    }
}

sub duration {
    my $self = shift;
    my $span = $self->{span} || die "Must set span!";
    $span =~ tr/'\\//d;
    $span = "$DAY_IN_SECONDS seconds" if $span eq 'today';
    
    my $dbh = $self->dbh;
    my $sql = <<__SQL__;
        SELECT EXTRACT(EPOCH FROM INTERVAL '$span');
__SQL__

    my $duration = $dbh->selectrow_array($sql);
    return abs($duration);
}

sub granularity {
    my $self = shift;
    my $full = shift;

    my $duration = $self->duration();

    my ($ival, $fmt);
    if ($duration <= $DAY_IN_SECONDS) {
        $ival = 'hour';
        $fmt  = 'HH24:MI';
    }
    elsif ($duration <= $MONTH_IN_SECONDS) {
        $ival = 'day';
        $fmt  = 'MM-DD';
    }
    elsif ($duration <= ($YEAR_IN_SECONDS * 2)) {
        $ival = 'month';
        $fmt  = 'YYYY-MM';
    }
    else {
        $ival = 'year';
        $fmt  = 'YYYY';
    }

    $fmt = 'YYYY-MM-DD HH24:MI' if $full;

    return ($ival, $fmt);
}

sub type_cases {
    my $self = shift;
    my $key  = shift;
    die "fields needs to be defined" unless $self->{fields};
    die "key needs to be defined" unless defined $key;

    my $cases = '';
    for my $field (@{$self->{fields}}) {
        $cases .= ",\n" if $cases;
        $cases .= sprintf("    COALESCE(SUM(%s),0) AS \"%s\"",
                          "CASE $key WHEN '$field' THEN tally END",
                          $field);
    }
    return $cases;
}

sub legend_items {
    my $self = shift;
    my $leg = $self->{legend} || [];
    return map {
	(my $col = $self->{graph}{dclrs}[$_] || '') =~ s/^#//;
	[ $col ? img({-src=>"/img-lib/legend_$col.gif"}) : (),
	  span({-class=>'body'}, $leg->[$_])
	];
    } 0 .. $#$leg;
}

sub legend_items_percent {
    my $self = shift;
    my $leg = $self->{legend} || [];

    my $data = $self->data || [];
    my $total = 0;
    $total += $_ for @{$data->[1]};

    return map {
	(my $col = $self->{graph}{dclrs}[$_] || '') =~ s/^#//;
	my $percent = sprintf('%.2f', $total ? ($data->[1][$_] / $total) * 100 : 0);
	[ $col ? img({-src=>"/img-lib/legend_$col.gif"}) : (),
	  span({-class=>'body'}, "$leg->[$_] ($percent%)")
	];
    } 0 .. $#$leg;
}

sub legend {
    my ($self,%opts) = @_;

    my $title = span({-class=>'subtitle'},"Legend");
    my $style = 'padding:0px; padding-right: 3px; padding-left: 0px; vertical-align: middle';

    my @cells = $opts{percent} ? $self->legend_items_percent : $self->legend_items;

    if ($opts{vertical}) {
	@cells = map { Tr( td({-style=>$style, -class=>'body'},$_->[0]),
	                   td({-style=>$style, -class=>'body'}, $_->[1])) 
		     } @cells;
	unshift @cells, Tr($title) unless $opts{notitle};
    }
    else {
	@cells = Tr(
	    $opts{notitle} ? () : td($title),
	    map { td({-style=>$style, -class=>'body'},$_->[0],$_->[1]) } @cells,
	);
    }

    return CGI::table({-class => 'dataTable',
	               -width => '90%',
		       -style => 'text-align:left; margin-left:auto; margin-right:auto',
		      },@cells);
}

sub table {
    my ($self,$data) = shift;
    my @table;

    $data ||= $self->data(type => 'table');

    return $self->nodata('table') unless $data;

    push @table, start_table({-class => 'dataTable',
	                      -width => '90%',
			      -style => 'margin-left:auto; margin-right:auto; text-align:left',
			     });

    push @table, Tr(td({-class => 'bodybold bordered'}, $self->{column_title}),
		    map { td({-class => 'body bordered'}, $_) } @{$data->[0]},
		   );

    my @legend = $self->legend_items;
    for my $i (1 .. $#$data) {
	my $class = $i % 2 ? 'body oddRow' : 'body evenRow';
	push @table, Tr(
	    td({-class => $class}, join(' ',@{$legend[$i-1]})),
	    map { td({-class => $class}, $_) } @{$data->[$i]},
	);
    }
    push @table, end_table;

    return div({-name  => 'dataTable',
	        -align => 'center',
		-class => 'dataTable',
	       },@table);
}

sub csv_data {
    my $self = shift;
    my $data = $self->data(csv => 1, @_) or return;
    return $data unless $self->{legend};
    return [
	[ $self->{column_title}, @{$data->[0]} ],
	map {[ $self->{legend}[$_-1], @{$data->[$_]} ]} 1 .. $#$data,
    ];
}

sub set_fonts {
    my ($self,$graph) = @_;

    my $font = "$PREFIX/var/www/fonts/ttf-bitstream-vera/Vera.ttf";

    $graph->set_title_font($font, $self->{font_size});
    $graph->set_values_font($font, $self->{font_size});
    $graph->set_x_axis_font($font, $self->{x_axis_font_size});
    $graph->set_y_axis_font($font, $self->{font_size});
    $graph->set_x_label_font($font, $self->{font_size});
    $graph->set_y_label_font($font, $self->{font_size});
}

sub image {
    my ($self,$imgtype) = @_;

    die "Must define imgtype" unless defined $imgtype;
    my $type = $self->{type} or die "no type!";
    my $data = $self->data;

    return $self->nodata($type,$imgtype) unless $data;

    my $graph = $self->new_graph($type);
    $graph->set( %{$self->{graph}}, %{$self->{$type}} ) or die $graph->error;
    $self->set_fonts($graph);
    my $gd = $graph->plot($data) || die $graph->error;
    return $self->new_image($gd,$imgtype);
}

sub new_image {
    my ($self,$gd,$imgtype) = @_;
    return $gd->$imgtype;
}

sub graph_class {
    my ($self,$type) = @_;
    die "no type!" unless $type;
    if ($type eq 'bars') {
	return 'Sophos::Reports::Graphic::GradientBars';
    }
    elsif ($type eq 'lines') {
	return 'Sophos::Reports::Graphic::Lines';
    }
    elsif ($type eq 'pie') {
	return 'Sophos::Reports::Graphic::GradientPie';
    }
    else {
	die "Unknown type: $type";
    }
}

sub new_graph {
    my ($self,$type) = @_;
    $type ||= $self->{type} || die "no type!";
    my $class = $self->graph_class($type);
    my $width = delete $self->{$type}{width} || $self->{width};
    my $height = delete $self->{$type}{height} || $self->{height};
    return $class->new($width, $height, 0);
}

sub nodata_img {
    my ($self,$type,$imgtype) = @_;
    $imgtype =~ s/^jpeg$/jpg/;

    my $nd_dir = "$PREFIX/$self->{nodata_dir}";
    my $nd_pth = $self->{nodata_path};

    if ($type eq 'table') {
	return "$nd_pth/$self->{nodata}_table.jpg";
    }
    else {
	return "$nd_dir/$self->{nodata}.$imgtype";
    }
}

sub nodata {
    my ($self,$type,@args) = @_;
    my $img = $self->nodata_img($type,@args);

    if ($type eq 'table') {
	return img({-src=>$img});
    }
    else {
	$self->print_binfile($img);
    }
}

sub print_binfile {
    my ($self,$binfile) = @_;
    die "$binfile not found" unless -f $binfile;
    local $/;
    open my $fh, $binfile or die "Can't open $binfile";
    binmode $fh;
    my $data = <$fh>;
    close $fh or die "Can't close $binfile";
    return $data;
}

sub masonize {
    my ($self, $file) = @_;

    my $outbuf;

    my $args = $self->{args} || {};
    my %args = %$args;
    $args{report} = $self;
 
    my $interp = HTML::Mason::Interp->new(out_method => \$outbuf);

    $interp->exec($file,%args);

    return $outbuf;
}

